#include <IOKit/IOLib.h>
#include <IOKit/platform/ApplePlatformExpert.h>
#include <IOKit/IODeviceTreeSupport.h>
#include <IOKit/IOLocks.h>
#include <IOKit/IOMessage.h>
#include <IOKit/pwr_mgt/RootDomain.h>
#include <IOKit/graphics/IOGraphicsInterfaceTypes.h>
#include <IOKit/i2c/IOI2CInterface.h>
#include <IOKit/pci/IOAGPDevice.h>
#include <IOKit/IOTimerEventSource.h>
#include <IOKit/assert.h>

#include <libkern/c++/OSContainers.h>
#include <string.h>

#include "ATY_Struct.h"

extern bool ATIPCIRangeIndexAlloc(IOService *driver, UInt8 index, UInt32 bar);
extern IOPhysicalAddress ATIPCIRangeMapGetPhysicalRange(IOService *driver, UInt8 index, UInt32 *bar, void** map);
extern IOVirtualAddress ATIPCIRangeMapAlloc(IOService *driver, UInt8 index, IOPhysicalAddress *addr);
extern IOVirtualAddress ATIPCIRangeMapAllocSubRange(IOService *driver, UInt8 index, IOPhysicalAddress offset, UInt32 length);
extern IOVirtualAddress ATIMapRegisters(IOService *driver);
extern IOVirtualAddress ATIMapRom(IOService *driver);
extern IOVirtualAddress ATIMapFrameBuffer(IOService *driver, IOPhysicalAddress offset, UInt32 length);
extern IOVirtualAddress ATIMapFrameBufferReserved(IOService *driver, IOPhysicalAddress offset, UInt32 length);
extern void ATIRequireMaxBusStall(IOService *service, UInt32 ns);

void MY_DEBUG(const char *fmt, ...) {
 va_list ap;
 va_start(ap, fmt);
 IOLog("ATY_HD: ");
 IOLog(fmt, ap);
}


OSStatus OPENDRVR(DriverGlobal *aDriverRecPtr, void* contents) {
	return noErr;
	/*
	OSStatus ret = noErr;
	MVAD mvad;
	DetailRecord* dRecord;
	CrtcValues *mode;
	UInt32 index;
	
	aDriverRecPtr->currentModeID = GetDefaultMode(aDriverRecPtr);
	//aDriverRecPtr->11622 = aDriverRecPtr->11621;
	//aDriverRecPtr->11623 = aDriverRecPtr->11622;
	aDriverRecPtr->currentDepth = kDepthMode3;
	bool done = false;
	if ((aDriverRecPtr->driverFlags & (1 << 1)) && !done) {
		done = true;
		if (aDriverRecPtr->dispNum == 0) {
			bcopy(aDriverRecPtr->mvad, &mvad, sizeof(MVAD));
			dRecord = aDriverRecPtr->dr;
		} else {
			bcopy(&aDriverRecPtr->mvad[1], &mvad, sizeof(MVAD));
			dRecord = &aDriverRecPtr->dr[1];
		}
		if (HALRestoreDetailTiming(aDriverRecPtr, mvad.modeID, dRecord) != noErr) done = 0;
	}
	if (done) {
		if (mvad.4 != aDriverRecPtr->monPara->0) done = false;
		if (aDriverRecPtr->edidP_lowByte != mvad.edidP_lowByte) done = false;
		if (aDriverRecPtr->driverFlags & (1 << 11)) done = false;	//CRT1?
	}
	if (done) {
		mode = HALFindCrtcValues(aDriverRecPtr, mvad.modeID);
		if (mode != NULL) {
			aDriverRecPtr->currentDepth = mvad.depth;
			aDriverRecPtr->currentModeID = mvad.modeID;
			if (IsDetailMode(aDriverRecPtr, mvad.modeID)) aDriverRecPtr->currentModeID = -5;
		}
	}
	mode = HALFindCrtcValues(aDriverRecPtr, aDriverRecPtr->currentModeID);
	if (mode == NULL) ret = controlErr;
	if (ret == noErr) {
		if (HALGetCrtcValues(aDriverRecPtr, aDriverRecPtr->currentModeID, aDriverRecPtr->currentDepth) == NULL)
			aDriverRecPtr->currentDepth = kDepthMode3;
		aDriverRecPtr->colorDepth = HALPixelSize(aDriverRecPtr, aDriverRecPtr->currentDepth, mode);
		aDriverRecPtr->colorBits = HALColorBits(aDriverRecPtr, aDriverRecPtr->currentDepth, mode);
		aDriverRecPtr->colorFormat = HALColorFormat(aDriverRecPtr, aDriverRecPtr->currentDepth, mode);
		aDriverRecPtr->gammaRecord = NULL;
		aDriverRecPtr->gammaRecord = PoolAllocateResident(sizeof(VDGammaRecord), 1);		//0x60C
		if ((aDriverRecPtr->gammaRecord == NULL) || (aDriverRecPtr->colorDepth == 0)) ret = controlErr;
	}
	aDriverRecPtr->aShare->driverFlags |= (1 << 4);
	some_data var_3C;
	if ((ret == noErr) && (aDriverRecPtr->driverFlags & (1 << 5))) {
		if (aDriverRecPtr->dispNum == 0) {
			ATIDeviceEnabler(var_3C, 5, aDriverRecPtr);
			ATIDeviceEnabler(var_3C, 0xA, aDriverRecPtr);
			index = 1;
		} else index = 8;
		ATIDeviceEnabler(var_3C, index, aDriverRecPtr);
		ATIDeviceEnabler(var_3C, 0x13, aDriverRecPtr);
		kdelay(1);
	}
	if (ret == noErr) {
		CreateLinearGamma(aDriverRecPtr);
		ret = HALSetupMyDisplay(aDriverRecPtr);
		HWProgramCRTCBlank(aDriverRecPtr, 0, aDriverRecPtr->dispNum);
	}
	aDriverRecPtr->driverFlags &= ~(1 << 4);
	if (reg == noErr) aDriverRecPtr->driverFlags |= 0x200000040;	//interrupt, newConnection
	return ret;
	 */
}

OSStatus CLOSEDRVR(DriverGlobal *aDriverRecPtr, void* contents) {
	return unimpErr;
	/*
	if (aDriverRecPtr->gammaRecord != NULL) {
		PoolDeallocate(aDriverRecPtr->gammaRecord);
		aDriverRecPtr->gammaRecord = NULL;
	}
	SetCrsrState(aDriverRecPtr, 0, 0, 0, aDriverRecPtr->dispNum);
	return noErr;
	 */
}

OSStatus CONTROLDRVR(DriverGlobal *aDriverRecPtr, void* contents) {
#ifdef ATY_HD_DEBUG
	const char cmdName[][35] = {
		"cscReset                       ",		//0
		"cscKillIO                      ",
		"cscSetMode                     ",
		"cscSetEntries                  ",
		"cscSetGamma                    ",
		"cscGrayPage                    ",
		"cscGrayScreen                  ",
		"cscSetGray                     ",
		"cscSetInterrupt                ",
		"cscDirectSetEntries            ",
		"cscSetDefaultMode              ",
		"cscSwitchMode                  ",
		"cscSetSync                     ",		//11
		"12                             ",
		"13                             ",
		"14                             ",
		"15                             ",
		"cscSavePreferredConfiguration  ",		//16
		"17                             ",
		"18                             ",
		"19                             ",
		"20                             ",
		"21                             ",
		"cscSetHardwareCursor           ",		//22
		"cscDrawHardwareCursor          ",
		"cscSetConvolution              ",
		"cscSetPowerState               ",
		"cscPrivateControlCall          ",		//26
		"27                             ",
		"cscSetMultiConnect             ",		//28
		"cscSetClutBehavior             ",		//29
		"30                             ",
		"cscSetDetailedTiming           ",		//31
		"32                             ",
		"cscDoCommunication             ",		//33
		"cscProbeConnection             ",		//34
		"35                             ",
		"cscSetScaler                   ",		//36
		"cscSetMirror                   ",
		"cscSetFeatureConfiguration     ",		//38
		"cscUnusedCall                  ",		//127
		"cscSetModeCallback             ",		//128
		"cscSetARMCallback              ",		//132
		"cscSetUnderscan                ",		//226
		"cscSetDispParams               "		//227
	};
#endif
	
	IONDRVControlParameters *pb = (IONDRVControlParameters *) contents;
	
	LOG1("%s	", cmdName[(pb->code < 39)?pb->code:((pb->code == 128)?40:((pb->code == 132)?41:((pb->code == 226)?42:((pb->code == 227)?43:39))))], false);
    switch (pb->code)
    {
        case cscSetEntries:
        case cscSetGamma:
            return noErr;
            break;
			
        default:
            return unimpErr;
            break;
    }
	
	/*
	if (!(aDriverRecPtr->driverFlags & 1)) return noErr;		//display not usable
	if (!(aDriverRecPtr->driverFlags & (1 << 12)) && (pb->code != cscSetPowerState)) return noErr;	//display in sleep
	if (pb->code > 227) return statusErr;						//227: cscSetDisplayParams
	switch (pb->code) {
		case cscReset:
			return DoVideoReset(pb->params, aDriverRecPtr);
		case cscSetMode:
			return DoSetVideoMode(pb->params, aDriverRecPtr);
		case cscSetEntries:
			return DoSetEntries(pb->params, aDriverRecPtr);
		case cscSetGamma:
			return DoSetGamma(pb->params, aDriverRecPtr);
		case cscGrayPage:
			return DoGrayPage(pb->params, aDriverRecPtr);
		case cscSetInterrupt:
			return DoSetInterrupt(pb->params, aDriverRecPtr);
		case cscDirectSetEntries:
			return DoDirectSetEntries(pb->params, aDriverRecPtr);
		case cscSetDefaultMode:
			return DoSetDefaultMode(pb->params, aDriverRecPtr);
		case cscSwitchMode:
			return DoSwitchMode(pb->params, aDriverRecPtr);
		case cscSetSync:
			return DoSetSync(pb->params, aDriverRecPtr);
		case cscSavePreferredConfiguration:
			return DoSaveConfiguration(pb->params, aDriverRecPtr);
		case cscSetHardwareCursor:
			return DoSetHardwareCursor(pb->params, aDriverRecPtr);
		case cscDrawHardwareCursor:
			return DoDrawHardwareCursor(pb->params, aDriverRecPtr);
		case cscSetPowerState:
			return DoSetPowerState(pb->params, aDriverRecPtr);
		case cscSetClutBehavior:
			return DoSetClutBehavior(pb->params, aDriverRecPtr);
		case cscSetDetailedTiming:
			return DoSetDetailTiming(pb->params, aDriverRecPtr);
		case cscDoCommunication:
			return DoATICommunication(pb->params, aDriverRecPtr);
		case cscProbeConnection:
			return DoSetVideoStatus(pb->params, aDriverRecPtr);
		case cscSetScaler:
			return DoSetScaler(pb->params, aDriverRecPtr);
		case cscSetMirror:
			return DoSetMirror(pb->params, aDriverRecPtr);
		case cscSetFeatureConfiguration:
			return DoSetFeatureConfiguratoin(pb->params, aDriverRecPtr);	//original source has the typo error
		case 128:		//this and below are not defined in IOMacOSVideo.h
			return DoATISetModeCallback(pb->params, aDriverRecPtr);
		case 132:
			return DoATISetARMCallback(pb->params, aDriverRecPtr);
		case 226:
			return DoSetUnderscan(pb->params, aDriverRecPtr);
		case 227:
			return DoATISetDispParams(pb->params, aDriverRecPtr);
		default:
			return controlErr;
	}
	 */
}

enum { kIOBootNDRVDisplayMode = 100 };

OSStatus STATUSDRVR(DriverGlobal *aDriverRecPtr, void* contents) {
#ifdef ATY_HD_DEBUG
	const char cmdName[][35] = {
		"0                              ",		//0
		"1                              ",
		"cscGetMode                     ",
		"cscGetEntries                  ",
		"cscGetPageCnt/cscGetPages      ",
		"cscGetPageBase/cscGetBaseAddr  ",
		"cscGetGray                     ",
		"cscGetInterrupt                ",
		"cscGetGamma                    ",
		"cscGetDefaultMode              ",
		"cscGetCurMode                  ",
		"cscGetSync                     ",
		"cscGetConnection               ",
		"cscGetModeTiming               ",
		"cscGetModeBaseAddress          ",
		"cscGetScanProc                 ",
		"cscGetPreferredConfiguration   ",
		"cscGetNextResolution           ",
		"cscGetVideoParameters          ",
		"19                             ",
		"cscGetGammaInfoList            ",
		"cscRetrieveGammaTable          ",
		"cscSupportsHardwareCursor      ",
		"cscGetHardwareCursorDrawState  ",
		"cscGetConvolution              ",
		"cscGetPowerState               ",
		"cscPrivateStatusCall           ",
		"cscGetDDCBlock                 ",
		"cscGetMultiConnect             ",
		"cscGetClutBehavior             ",
		"cscGetTimingRanges             ",
		"cscGetDetailedTiming           ",
		"cscGetCommunicationInfo        ",
		"33                             ",
		"cscGetVideoStatus              ",
		"cscGetScalerInfo               ",
		"cscGetScaler                   ",
		"cscGetMirror                   ",
		"cscGetFeatureConfiguration     ",		//38
		"cscGetFeatureList              ",		//39
		"cscUnusedCall                  ",		//127
		"cscATIGetInfo                  ",		//128
		"cscGetUnderscan                ",		//226
		"cscGetDispParams               ",		//227
	};
#endif

	IONDRVControlParameters* pb = (IONDRVControlParameters*) contents;
	OSStatus ret;

	LOG1("%s	", cmdName[(pb->code < 40)?pb->code:((pb->code == 128)?41:((pb->code == 226)?42:((pb->code == 227)?43:40)))], false);
    switch (pb->code)
    {
        case cscGetCurMode:
		{
			VDSwitchInfoRec * switchInfo = (VDSwitchInfoRec *) (pb->params);
			
			switchInfo->csData     = kIOBootNDRVDisplayMode;
			switchInfo->csMode     = kDepthMode1;
			switchInfo->csPage     = 1;
			switchInfo->csBaseAddr = (Ptr) (1 | (uintptr_t) (aDriverRecPtr->baseAddr));
			ret = noErr;
		}
            break;
			
        case cscGetNextResolution:
		{
			VDResolutionInfoRec * resInfo = (VDResolutionInfoRec *) (pb->params);
			
			if ((kDisplayModeIDFindFirstResolution == (SInt32) resInfo->csPreviousDisplayModeID)
				|| (kDisplayModeIDCurrent == (SInt32) resInfo->csPreviousDisplayModeID))
			{
				resInfo->csDisplayModeID 	= kIOBootNDRVDisplayMode;
				resInfo->csMaxDepthMode		= kDepthMode1;
				resInfo->csHorizontalPixels	= aDriverRecPtr->width;
				resInfo->csVerticalLines	= aDriverRecPtr->height;
				resInfo->csRefreshRate		= 0 << 16;
				ret = noErr;
			}
			else if (kIOBootNDRVDisplayMode == resInfo->csPreviousDisplayModeID)
			{
				resInfo->csDisplayModeID = kDisplayModeIDNoMoreResolutions;
				ret = noErr;
			}
			else
			{
				resInfo->csDisplayModeID = kDisplayModeIDInvalid;
				ret = paramErr;
			}
		}
            break;
			
        case cscGetVideoParameters:
		{
			VDVideoParametersInfoRec * pixelParams = (VDVideoParametersInfoRec *) (pb->params);
			
			if ((kIOBootNDRVDisplayMode != pixelParams->csDisplayModeID)
				|| (kDepthMode1 != pixelParams->csDepthMode))
			{
				ret = paramErr;
				break;
			}
			VPBlock *	pixelInfo = pixelParams->csVPBlockPtr;
			
			pixelInfo->vpBounds.left	= 0;
			pixelInfo->vpBounds.top	= 0;
			pixelInfo->vpBounds.right	= aDriverRecPtr->width;
			pixelInfo->vpBounds.bottom	= aDriverRecPtr->height;
			pixelInfo->vpRowBytes	= aDriverRecPtr->pitch;
			pixelInfo->vpPlaneBytes	= 0;
			pixelInfo->vpPixelSize	= aDriverRecPtr->colorDepth;
			ret = noErr;
		}
            break;
			
        case cscGetModeTiming:
		{
			VDTimingInfoRec * timingInfo = (VDTimingInfoRec *) (pb->params);
			
			if (kIOBootNDRVDisplayMode != timingInfo->csTimingMode)
			{
				ret = paramErr;
				break;
			}
			timingInfo->csTimingFormat = kDeclROMtables;
			timingInfo->csTimingFlags  = kDisplayModeValidFlag | kDisplayModeSafeFlag;
			ret = noErr;
		}
            break;
			
        default:
            ret = unimpErr;
            break;
    }
	
	return ret;
	
	/*
	if (!(aDriverRecPtr->driverFlags & 1)) return noErr;		//display not usable
	if (!(aDriverRecPtr->driverFlags & (1 << 12)) && (pb->code != cscGetPowerState))
		return noErr;											//display in sleep
	if (pb->code > 227) return statusErr;						//227: cscGetDisplayParams
	switch (pb->code) {
		case cscGetMode:
			return DoGetMode(pb->params, aDriverRecPtr);
		case cscGetEntries:
			return DoGetEntries(pb->params, aDriverRecPtr);
		case cscGetPages:
			return DoGetPages(pb->params, aDriverRecPtr);
		case cscGetBaseAddr:
			return DoGetBaseAddr(pb->params, aDriverRecPtr);
		case cscGetInterrupt:
			return DoGetInterrupt(pb->params, aDriverRecPtr);
		case cscGetGamma:
			return DoGetGamma(pb->params, aDriverRecPtr);
		case cscGetDefaultMode:
			return DoGetDefaultMode(pb->params, aDriverRecPtr);
		case cscGetCurMode:
			return DoGetCurrentMode(pb->params, aDriverRecPtr);
		case cscGetSync:
			return DoGetSync(pb->params, aDriverRecPtr);
		case cscGetConnection:
			return DoGetConnection(pb->params, aDriverRecPtr);
		case cscGetModeTiming:
			return DoGetModeTiming(pb->params, aDriverRecPtr);
		case cscGetPreferredConfiguration:
			return DoGetConfiguration(pb->params, aDriverRecPtr);
		case cscGetNextResolution:
			return DoGetNextResolution(pb->params, aDriverRecPtr);
		case cscGetVideoParameters:
			return DoGetVideoParameters(pb->params, aDriverRecPtr);
		case cscSupportsHardwareCursor:
			return DoSupportsHardwareCursor(pb->params, aDriverRecPtr);
		case cscGetHardwareCursorDrawState:
			return DoGetHardwareCursorState(pb->params, aDriverRecPtr);
		case cscGetPowerState:
			return DoGetPowerState(pb->params, aDriverRecPtr);
		case cscGetDDCBlock:
			return DoFetchDDCBlock(pb->params, aDriverRecPtr);
		case cscGetClutBehavior:
			return DoGetClutBehavior(pb->params, aDriverRecPtr);
		case cscGetTimingRanges:
			return DoGetTimingRanges(pb->params, aDriverRecPtr);
		case cscGetDetailedTiming:
			return DoGetDetailTiming(pb->params, aDriverRecPtr);
		case cscGetCommunicationInfo:
			return DoATIGetCommunicationInfo(pb->params, aDriverRecPtr);
		case 34:		//not defined in IOMacOSVideo.h
			return DoGetVideoStatus(pb->params, aDriverRecPtr);
		case cscGetScalerInfo:
			return DoGetScalerInfo(pb->params, aDriverRecPtr);
		case cscGetScaler:
			return DoGetScaler(pb->params, aDriverRecPtr);
		case cscGetMirror:
			return DoGetMirror(pb->params, aDriverRecPtr);
		case cscGetFeatureConfiguration:
			return DoGetFeatureConfiguratoin(pb->params, aDriverRecPtr);	//original source has the typo error
		case 39:		//this and below are not defined in IOMacOSVideo.h
			return DoGetFeatureList(pb->params, aDriverRecPtr);
		case 128:
			return DoATIGetInfo(pb->params, aDriverRecPtr);
		case 226:
			return DoGetUnderscan(pb->params, aDriverRecPtr);
		case 227:
			return DoATIGetDispParams(pb->params, aDriverRecPtr);
		default:
			return statusErr;
	}
	 */
}

OSStatus DoOpenCmd(DriverGlobal *aDriverRecPtr, void* contents) {
	aDriverRecPtr->instCount++;
	if (aDriverRecPtr->instCount == 1) return OPENDRVR(aDriverRecPtr, contents);
	return noErr;
}

OSStatus DoCloseCmd(DriverGlobal *aDriverRecPtr, void* contents) {
	aDriverRecPtr->instCount--;
	if (aDriverRecPtr->instCount == 0) return CLOSEDRVR(aDriverRecPtr, contents);
	return noErr;
}

OSStatus DoKillIOCmd(DriverGlobal *aDriverRecPtr, void* contents) {
	return noErr;
}

OSStatus InternalStart(DriverGlobal *aDriverRecPtr, UInt16 refNum, RegEntryID regID, bool isReplace) {
	PE_Video		bootDisplay;
	UInt32			bpp;
	IODeviceMemory	* mem;
	UInt32          numMaps, i;
	bool            matched = false;
	IOService		*service;

	if (isReplace) return unimpErr;
	if (aDriverRecPtr == NULL) return controlErr;
	if (aDriverRecPtr->aProvider == NULL) return controlErr;
	service = aDriverRecPtr->aProvider;
	if (service == NULL) return controlErr;
	
	IOService::getPlatform()->getConsoleInfo( &bootDisplay);
	numMaps = service->getDeviceMemoryCount();
	for (i = 0; (!matched) && (i < numMaps); i++)
	{
	    mem = service->getDeviceMemoryWithIndex(i);
	    if (!mem) continue;
	    matched = (bootDisplay.v_baseAddr >= mem->getPhysicalAddress())
		&& ((bootDisplay.v_baseAddr < (mem->getPhysicalAddress() + mem->getLength())));
	}
		
	if (matched)
	{
	    aDriverRecPtr->baseAddr	    = (void *) bootDisplay.v_baseAddr;
	    aDriverRecPtr->pitch	    = bootDisplay.v_rowBytes;
	    aDriverRecPtr->width	    = bootDisplay.v_width;
	    aDriverRecPtr->height	    = bootDisplay.v_height;
	    bpp = bootDisplay.v_depth;
	    if (bpp == 15)
			bpp = 16;
	    else if (bpp == 24)
			bpp = 32;
	    aDriverRecPtr->colorDepth = bpp;
		return noErr;
	}
	return controlErr;

	/*
	OSStatus ret = noErr;
	if (aDriverRecPtr == NULL) return noBridgeErr;
	int i;
	for (i = 0;i < 4;i++) aDriverRecPtr->regIDSelf.opaque[i] = regID.opaque[i];
	aDriverRecPtr->refNum = refNum;
	RegPrint(&aDriverRecPtr->regIDSelf, "ndrv_version", "1.1f39", 7);
	RegPrint(&aDriverRecPtr->regIDSelf, "mach_o", "1.1f39", 7);
	RegPrint(&aDriverRecPtr->regIDSelf, "ndrv_date", "Aug 16 2006", 12);
	ret = InitPCIDevice(aDriverRecPtr);
	if (ret != noErr) return ret;
	
	if (!aDriverRecPtr->D302) {
		aDriverRecPtr->dispNum = 0;
		if (!RegistryEntryIDCompare(&aDriverRecPtr->regIDParent, &aDriverRecPtr->regIDSelf))
			aDriverRecPtr->dispNum = FindDisplayNumber(aDriverRecPtr, &aDriverRecPtr->regIDSelf, &aDriverRecPtr->regIDParent);
	}
	
	if (aDriverRecPtr->connectNum < aDriverRecPtr->dispNum) return controlErr;
	
	if (aDriverRecPtr->dispNum == 0) {
		aDriverRecPtr->aShare = PoolAllocateResident(sizeof(Share_Global), 1);
		if (!aDriverRecPtr->aShare) return controlErr;
		RegPrint(&aDriverRecPtr->regIDParent, "my_shared_globals", aDriverRecPtr->aShare, 4);
	} else
		RegGet(&aDriverRecPtr->regIDParent, "my_shared_globals", aDriverRecPtr->aShare, 4);
	
	if (aDriverRecPtr->dispNum == 0)
		RegPrint(&aDriverRecPtr->regIDParent, "my_Id1", aDriverRecPtr->regIDSelf, 16);
	else
		RegPrint(&aDriverRecPtr->regIDParent, "my_Id2", aDriverRecPtr->regIDSelf, 16);
	
	if (ret == noErr) ret = InitTimingTable(aDriverRecPtr);
	if (ret == noErr) ret = InitTVParameters(aDriverRecPtr);
	if (ret == noErr) InitPreferences(aDriverRecPtr);
	if (ret == noErr) ret = InitAddresses(aDriverRecPtr, aDriverRecPtr->dispNum);
	if (ret == noErr) ret = InitHardware(aDriverRecPtr);
	if (ret == noErr) ret = InitConfiguration(aDriverRecPtr);
	if (ret == noErr) ret = InitInterrupts(aDriverRecPtr);
	
	if (ret == noErr) {
		aDriverRecPtr->driverFlags |= 1;
		UInt32 data = aDriverRecPtr->dispNum + 1;
		RegPrint(&aDriverRecPtr->regIDParent, "my_count", &data, 4);
		if (aDriverRecPtr->dispNum != 0) {
			data = 0x73656364;		//"secd"
			RegPrint(&aDriverRecPtr->regIDSelf, "my_display", &data, 4);
		} else {
			data = 0x7072696D;		//"prim"
			RegPrint(&aDriverRecPtr->regIDSelf, "my_display", &data, 4);
		}
		InitBootOptions(aDriverRecPtr);
	}
	
	return ret; */
}

OSStatus InternalEnd(DriverGlobal *aDriverRecPtr, UInt16 refNum, RegEntryID regID, bool isReplace) {
	return unimpErr;
	/*
	if (aDriverRecPtr == NULL) return noErr;
	
	UInt32 data;
	if (aDriverRecPtr->driverFlags & 1) {
		RegGet(&aDriverRecPtr->regIDParent, "my_count", &data, 4);
		if (data) data--;
		RegPrint(&aDriverRecPtr->regIDParent, "my_count", &data, 4);
	}
	
	if 	(isReplace) {
		Save_DR drSave;
		drSave.depth = aDriverRecPtr->currentDepth;
		drSave.modeID = aDriverRecPtr->currentModeID;
		drSave.4 = aDriverRecPtr->monPara->0;
		drSave.edidP_lowByte = aDriverRecPtr->edidP_lowByte;
		drSave.3 = 0;
		HALSaveDetailRecord(aDriverRecPtr, &drSave.dr, &drSave, aDriverRecPtr->currentModeID, 0);
		drSave.28 = aDriverRecPtr->4A8;
		drSave.serviceID = aDriverRecPtr->serviceID;
		drSave.20 = aDriverRecPtr->8;
		DoReplacementInfo(aDriverRecPtr, &drSave, 1);
	} else {
		aDriverRecPtr->driverFlags &= ~1;		//disable the driver
		DeviceOutputsOff(aDriverRecPtr);
		HWProgramCRTCOff(aDriverRecPtr, aDriverRecPtr->dispNum);
	}
	
	FreeInterrupts(aDriverRecPtr);
	return noErr;
	 */
}

OSStatus DoInitializeCmd(DriverGlobal *aDriverRecPtr, void *contents) {
	DriverInitInfo *info = (DriverInitInfo *)contents;
	return InternalStart(aDriverRecPtr, info->refNum, info->deviceEntry, false);
}

OSStatus DoReplaceCmd(DriverGlobal *aDriverRecPtr, void* contents) {
	DriverInitInfo *info = (DriverInitInfo *)contents;
	return InternalStart(aDriverRecPtr, info->refNum, info->deviceEntry, true);
}

OSStatus DoFinalizeCmd(DriverGlobal *aDriverRecPtr, void* contents) {
	DriverInitInfo *info = (DriverInitInfo *)contents;
	return InternalEnd(aDriverRecPtr, info->refNum, info->deviceEntry, false); 
}

OSStatus DoSupersededCmd(DriverGlobal *aDriverRecPtr, void* contents) {
	DriverInitInfo *info = (DriverInitInfo *)contents;
	return InternalEnd(aDriverRecPtr, info->refNum, info->deviceEntry, true); 
}

OSStatus DoReadCmd(DriverGlobal *aDriverRecPtr, void* contents) {
	return noErr;
}

OSStatus DoWriteCmd(DriverGlobal *aDriverRecPtr, void* contents) {
	return noErr;
}

OSStatus DoControlCmd(DriverGlobal *aDriverRecPtr, void* contents) {
	return CONTROLDRVR(aDriverRecPtr, contents);
}

OSStatus DoStatusCmd(DriverGlobal *aDriverRecPtr, void* contents) {
	return STATUSDRVR(aDriverRecPtr, contents);
}

OSStatus DoDriverIO(DriverGlobal *aDriverRecPtr, int index, UInt32 cmdID,
			 		void *contents, UInt32 cmdCode, UInt32 cmdKind) {
	OSStatus ret = noErr;
	//if (aDriverRecPtr != NULL) aDriverRecPtr->8 = index;	//always 0
	if ((cmdCode - kIONDRVOpenCommand) < 11) {
	  switch (cmdCode) {
		case kIONDRVInitializeCommand:
			return DoInitializeCmd(aDriverRecPtr, contents);
		case kIONDRVFinalizeCommand:
			return DoFinalizeCmd(aDriverRecPtr, contents);
		case kIONDRVReplaceCommand:
			return DoReplaceCmd(aDriverRecPtr, contents);
		case kIONDRVSupersededCommand:
			return DoSupersededCmd(aDriverRecPtr, contents);
		case kIONDRVOpenCommand:
			return DoOpenCmd(aDriverRecPtr, contents);
		case kIONDRVCloseCommand:
			return DoCloseCmd(aDriverRecPtr, contents);
		case kIONDRVKillIOCommand:
			return DoKillIOCmd(aDriverRecPtr, contents);
		case kIONDRVReadCommand:
			ret = DoReadCmd(aDriverRecPtr, contents);
			break;
		case kIONDRVWriteCommand:
			ret = DoWriteCmd(aDriverRecPtr, contents);
			break;
		case kIONDRVControlCommand:
			ret = DoControlCmd(aDriverRecPtr, contents);
			break;
		case kIONDRVStatusCommand:
			ret = DoStatusCmd(aDriverRecPtr, contents);
			break;
		default:
			break;
	  }
	}
	//if (!(cmdKind & kIONDRVImmediateIOCommandKind)) ret = IOCommandIsComplete((OpaqueIOCommandID *) cmdID, ret);
	return ret;
}

