#ifndef _ATY_STRUCT_H
#define _ATY_STRUCT_H

#include <IOKit/ndrvsupport/IONDRVFramebuffer.h>
#include "IONDRV.h"
#include "IONDRVLibraries.h"

#define ATY_HD_DEBUG

/* __func__ is really nice, but not universal */
#if !defined(__GNUC__) && !defined(C99)
#define __func__ "unknown"
#endif

#ifdef ATY_HD_DEBUG
#define RHDFUNC(ptr) IOLog("ATY_HD: %s\n", __func__)
#define RHDFUNCI(scrnIndex) IOLog("ATY_HD: %s\n", __func__)
#define LOG(fmt, h) do {if ((h)) IOLog("ATY_HD: "); IOLog((fmt));} while (0)
#define LOG1(fmt, a1, h) do {if ((h)) IOLog("ATY_HD: "); IOLog((fmt), (a1));} while (0)
#define LOG2(fmt, a1, a2, h) do {if ((h)) IOLog("ATY_HD: "); IOLog((fmt), (a1), (a2));} while (0)
#define LOG3(fmt, a1, a2, a3, h) do {if ((h)) IOLog("ATY_HD: "); IOLog((fmt), (a1), (a2), (a3));} while (0)
void MY_DEBUG(const char *fmt, ...);	//this seems not working
#else
#define RHDFUNC(ptr)
#define RHDFUNCI(scrnIndex)
#define LOG(fmt)
#define LOG1(fmt, a1)
#define LOG2(fmt, a1, a2)
#define LOG3(fmt, a1, a2, a3)
inline void MY_DEBUG(const char *fmt, ...) {}
#endif

typedef struct RegEntryID RegEntryID;
typedef struct DriverInitInfo DriverInitInfo;

enum {
	noBridgeErr					= -93,
};

typedef struct {
	UInt32				modeID;
	UInt32				mFlag;				//bit2 == 1 means default mode
} ModeFlag;

typedef struct {
	UInt32				minID;
	UInt32				maxID;
	UInt8				tFlag;
} TimingFlags;

typedef struct {
	UInt32				atyMemSize;
	UInt32				vramMemSize;
} ATYVRAM;

typedef struct {
	UInt32				size;				//0
											//4
											//8
	UInt32				totalVram;			//C
	IOVirtualAddress	IOMapBase;			//1C
	IOVirtualAddress	screenOffset;		//24
	UInt32				chipID;				//48
} ATIInfo;

typedef struct {
	UInt32				aliasModeID;		//0
	UInt32				modeID;				//4
} AliasIDs;

typedef struct {
	UInt32				modeID;				//0
	UInt32				timingMode;			//4
	ushort				mFlag;				//8 0:Horizontal Polarity, 1:Vertical Polarity, 3:interlaced, 4:DFP2
												//6:stretch, 4:onOff, 8&9:rotate90, 9&10:rotate180, 8&10:rotate270
												//2:hSyncACntl, 
	ushort				Clock;				//A 10000Hz as unit
	ushort				width;				//C
	ushort				HTotal;				//E
	ushort				dHSyncS;			//10
	ushort				dHSyncE;			//12
	ushort				height;				//14
	ushort				VTotal;				//16 525, 625, 750, 1125
	ushort				dVSyncS;			//18
	ushort				dVSyncE;			//1A
} CrtcValues;

typedef struct {
	UInt32				modeID;				//0
	UInt32				timingMode;			//4
	UInt32				mFlag;				//8
	UInt32				Clock;				//C 10000Hz as unit
	UInt32				width;				//10
	UInt32				HTotal;				//14
	UInt32				dHSyncS;			//18
	UInt32				dHSyncE;			//1C
	UInt32				height;				//20
	UInt32				VTotal;				//24
	UInt32				dVSyncS;			//28
	UInt32				dVSyncE;			//2C
} CrtcValuesV2;

typedef struct {
	ushort				width;				//0
	ushort				height;				//2
} ScaledParameters;

typedef struct {
	UInt32				modeID;				//0
	UInt32				modeSeed;			//4
	UInt32				modeState;			//8
	UInt32				modeAlias;			//C
	UInt32				Clock;				//10
	UInt32				width;				//14
	UInt32				height;				//18
	UInt32				horiInset;			//1C
	UInt32				verInset;			//20
} DetailTimingState;

typedef struct {							//size 0x18
	ushort				clock;				//0
	ushort				width;				//2
	ushort				HTotal;				//4
	ushort				dHSyncS;			//6
	ushort				dHSyncE;			//8
	ushort				height;				//A
	ushort				VTotal;				//C
	ushort				dVSyncS;			//E
	ushort				dVSyncE;			//10
	ushort				width2;				//12
	ushort				height2;			//14
	UInt8				flagHB;				//16
	UInt8				flagLB;				//17
} DetailRecord;

typedef struct {
	ushort				unknown1;
	UInt8				unknown2;
	UInt8				unknown3;
	ushort				unknown4;
} MonitorParameter;

typedef struct {
	UInt32				Clock;				//0
	UInt32				maxHFreq;			//4
	UInt32				minHFreq;			//8
	UInt32				maxRefRate;			//C
	UInt32				minRefRate;			//10
	UInt32				maxWidth;			//14
	UInt32				maxHeight;			//18
} monitorLimit;

typedef struct {
	UInt8				crtcID;				//0
	UInt8				connectedFlags;		//1
	UInt8				connectTaggedType;	//2
	UInt8				unknown1;			//3
	UInt8				connectTaggedData;	//4
	UInt8				unknown2;			//5
	UInt8				unknown3;			//6
	UInt8				unknown4;			//7
	UInt32				edidP1;				//8
	UInt32				edidP2;				//C
	UInt8				unknown5;			//10
	UInt8				edidFlags;			//11
} ConnectionInfo;

typedef struct {
	UInt8				**features;
	UInt32				size;
} FeatureList;

typedef struct {
	UInt32				array[2];			//0
	UInt32				flags[2];			//28
	UInt32				rDegree[2];			//30
} ATIFEDSInfo;

typedef struct {							//size 0x260
	UInt32				size;				//0 = 0x260 always
	UInt32				unknown1;			//4 = 0x44503031/'DP01'
	UInt32				unknown2;			//8
	UInt8				connectedFlags;		//C
	UInt8				controlFlags;		//10
	UInt8				activeFlags;		//14
	UInt8				timingFlags;		//18
	ushort				scaledFlags;		//1C
	UInt8				crtcID;				//20
	UInt32				connectTaggedData;	//24
											//28
											//2A
											//30
											//32
											//34
											//36
	ushort				unknown3;			//38
											//3A
	UInt8				unknown4;			//3C
											//3D
											//3E
											//41
											//44	
	UInt8				unknown5;			//47
											//4A
	UInt8				unknown6;			//51
	UInt8				unknown7;			//56
	UInt8				unknown8;			//57
											//5A
	UInt8				unknown9;			//5D
	UInt8				unknown10;			//5E
	UInt8				unknown11;			//5F
	UInt8				model[256];			//60
	UInt8				slotName[256];		//160
} DisplayParameters;

typedef struct {							//size 0x4000
	
} CursorImage;

typedef struct {
	short				gVersion;
	short				gType;
	short				gFormulaSize;		//Formula data size
	short				gChanCnt;			//1 or 3
	short				gDataCnt;			//data count of each table
	short				gDataWidth;			//data bits
	UInt32				gFormulaData[1];	//point to entry of Formula data & Correction data (gamma table)
} GammaTb1;

typedef struct {
	UInt32				configMemSize;			//0 = regr32(0xF8)
	UInt32				agpLocation;			//4
	UInt32				cmdPadCntl1;			//8
	UInt32				cmdPadCntl2;			//C
												//20
												//24
	UInt32				unknown1;				//28
	UInt32				writeAge1;				//2C
	UInt32				writeAge2;				//30
	UInt32				dataPadCntl1;			//34
	UInt32				dataPadCntl2;			//38
	UInt32				agpBase1;				//3C
												//12C
	UInt32				unknown2;				//138
	UInt32				memResetSeqNum;			//13C
	UInt32				memResetSequence[80];	//140 0x80 is not clear yet
	UInt32				atySettings[13];		//340
} MemoryConfiguration;

typedef struct {
	short				modeID;					//0
	short				depth;					//2
	long				unknown1;				//4
	UInt8				unknown2;				//7
} MVAD;

typedef struct {
	ushort				depth;					//0
	short				modeID;					//2
												//8
} ATIConfigurationRec;

typedef struct {								//size 0x654
	UInt32				displayNum;				//0
	UInt32				memSize;				//8 = regr32(0xF8)
	UInt32				totalVram;				//C
												//10
	RegEntryID			mirrorEntryID[2];		//14 = ndrv, 18 = ?
	ConnectionInfo		connectionInfo;			//1C
	UInt8				array[2];				//63 = crtcID
	UInt32				array2[2];				//6C, 70
												//80
												//84
	struct _VSLService	*serviceID[2];			//98
	MemoryConfiguration	memConfig;				//A0
	UInt32				memClock;				//414
	UInt32				sysClock;				//418
	UInt32				clock1;					//41C = 414
	UInt32				clock2;					//420 = 418
	UInt32				driverFlags;			//42C
	UInt8				_crtcID;				//430
												//431
	UInt8				crtcID;					//433
	UInt32				unknown1;				//434
	UInt32				unknown2;				//438
	UInt32				unknown3;				//43C
	UInt32				unknown4;				//448
	UInt32				unknown5;				//44C
	UInt32				unknown6;				//450
	AbsoluteTime		startTime;				//474, size 8
	UInt32				iconColor1[2];			//490
	UInt32				iconColor2[2];			//498
	UInt32				array3[2];				//4A0
	UInt32				array4[2];				//4A8, size 0x44
	UInt32				iconStart[2];			//530
	UInt32				array5[2];				//538
	UInt32				array6[2];				//540
	UInt32				iconSizeX[2];			//548
	UInt32				iconSizeY[2];			//550
	UInt32				array7[2];				//558
	UInt32				array8[2];				//560
	UInt32				array9[2];				//568
	UInt32				array10[2];				//570
	UInt32				array11[2];				//578
	UInt32				array12[2];				//580,584
	UInt8				array13[2];				//588
	UInt8				array14[2];				//58A
	UInt32				cursorPosition;			//628
	UInt8				unknown7;				//62C
	struct DriverGlobal	*aDriverRecPtrs[2];		//6A0
	//ATYSurfaceInfo		surfaceInfo;			//64C
} ATYSG;	//shared global in IORegistry

typedef struct DriverGlobal {					//size 0x11794
	UInt32				instCount;				//0
	UInt32				driverFlags;			//4:	24 HWScaler, 25 SWScaler, 26 RotationScaler,
														//0 on/off, 18 noHotPlug, 23 noHPInterrupt, 2 atyFlags available
														//11 CRT1/DDCBlock, 13 HWMirror, 8 ClutBehavior,
														//29 interrupt, 12 wake, 28 something changed or not,
														//16:HWScaler, 15:rotate/swscale, 6:newConnection, 5:interrupt
	UInt32				index;					//8 always 0
	RegEntryID			regIDSelf;				//C
	RegEntryID			regIDparent;			//1C
	RegEntryID			atiRegs[2];				//2C, each size 0x10
												//4C
	MVAD				*mvad1;					//4E size 0x40
	UInt8				unknown1;				//50
	UInt32				unknown2;				//52
	MVAD				*mvad2;					//56 size 0x40
	UInt8				unknown3;				//58
	DetailRecord		*dr1;					//5E size 24
	UInt32				unknown4;				//5A
	UInt32				unknown5;				//6A
	DetailRecord		*dr2;					//76 size 24
	UInt16				refNum;					//84
												//90
	UInt8				unknown6;				//94 = 95
	UInt8				connectionNum;			//95
	UInt8				revID;					//96
	UInt32				chipID;					//98 including DevID and VenID
	UInt32				totalVram;				//9C
	UInt8				unknown7;				//A0
	IOPhysicalAddress	aaplIOBase;				//A4 kIOPCIConfigBaseAddress2
	IOPhysicalAddress	aaplExpRomBase;			//A8 kIOPCIConfigExpansionROMBase
	IOPhysicalAddress	aaplVramBase;			//AC kIOPCIConfigBaseAddress0
	void				*aaplIOMap;				//B0
	void				*aaplExpRomMap;			//B4
	void				*aaplVramMap;			//B8
	IOVirtualAddress	FBMapBase;				//BC
	IOVirtualAddress	IOMapBase;				//C4
	IOVirtualAddress	ExpRomMapBase;			//C8
	IOVirtualAddress	screenMapBase;			//CC
	UInt32				screenOffset;			//D0 = 115B0->C - 9C
	IOPhysicalAddress	IOPhysBase;				//D4
	IOPhysicalAddress	ExpRomPhysBase;			//D8
	UInt32				screenSize;				//DC
	void				*baseAddr;				//E0
	IOVirtualAddress	currentFBMapBase;		//E4
	UInt32				subFBLength;			//E8 initialized as 0x20000 and never changed
	MemoryConfiguration	*memConfig;				//EC = 115B0->A0
	UInt8				unknown8;				//F0
	UInt32				unknown9;				//F4
	struct _VSLService	*serviceID;				//F8
												//FC
	UInt32				powerState;				//110 these two relate to powerState
	UInt32				powerFlags;				//114
	UInt32				currentModeID;			//118
	UInt32				currentDepth;			//11C default as kDepthMode3 {8, 32}
	UInt32				colorDepth;				//120 bitsPerPixel
	UInt32				colorFormat;			//124 8,16,32:0 64:0,1,2,3
	UInt32				colorBits;				//128 bitsPerColor 8:8 16:5 32:10(ARGB2101010),8(others) 64:16
	UInt32				pitch;					//12C
	UInt32				width;					//130
	UInt32				height;					//134
	UInt32				deltaHOverScan;			//138
	UInt32				deltaVOverScan;			//13C
	CrtcValues			*currentMode;			//140
	UInt32				changeFlags;			//144 0:mode, 1:pitch, 14:screenOffset/Size, 3:driverFlags bit15, 7:->1152C/width
														//6:->1152E/height, 1&4:->4bit15, 4:->11534/11536/->144bit8==1,
														//2:timingFlags, 9:depth, 5:rotate, 8:width/height, 7:->1152C
					 
	MonitorParameter	*monPara;				//148
	UInt32				modesNum;				//14C
	UInt32				modesIndex;				//150
	//ModeFlag			*modeFlags[maxModesNum];//154 each size 8
	//CrtcValues			*modes[maxModesNum];	//158 each size 0x1C
	UInt32				maxModesNum;			//15C
	UInt32				aliasIDNum;				//160
	//AliasIDs			*aliasID[maxModesNum];	//164 each size 8
	UInt8				underScan;				//168
	UInt8				DTileMode;				//169
	//					(*DDCSetSense)();		//16C, ddcSetData/Clock callback
	//					(*DDCGetSense)();		//170, ddcGetData/Clock callback
	VDGammaRecord		*gammaRecord;			//174
	UInt32				colorRange[256];		//178
	UInt8				unknown10;				//430
	UInt8				unknown11;				//431
	UInt32				unknown12;				//578
	UInt32				unknown13;				//57C
	UInt32				ARM[5];					//580-590
												//594
												//598
												//5A4
	CursorImage			crsrImage1;				//5A8  size 0x4000, 64*64 at biginning, crsrImage data
	CursorImage			crsrImage2;				//45A8 size 0x4000, 64*64, crsrImage data
	CursorImage			crsrImage3;				//85A8 size 0x4000, 64*64, crsrImage data
	CursorImage			crsrImage4;				//C5A8 size 0x4000, 64*64, crsrImage data
	UInt32				cursorAddress;			//105A8
	UInt32				cursorX;				//105AC
	UInt32				cursorY;				//105B0
	UInt32				unknown14;				//105B4
	UInt32				unknown15;				//105B8
	Boolean				cursorVisible;			//105BC
	UInt8				cursorMode;				//105C0
	UInt8				unknown16;				//105C1
	Boolean				cursorSet;				//105C2
	UInt32				iconSurfaceAddr;		//105C4
	UInt32				unknown17;				//105D0
	UInt8				unknown18;				//105D4
												//105D8
	UInt8				sync;					//105D5 0:hsync, 1:vSync, 2:both, 7:sync
	RGBColor			rgbData[2];				//105E0 each size 8
	ColorSpec			colorTable[256];		//105F0 each size 8
	DisplayParameters	previousDispPara;		//10DF0 size 0x260
	DisplayParameters	currentDispPara;		//11050 size 0x260
	UInt32				atyFlags;				//112B0
	UInt8				unknown19;				//112B4
	UInt8				unknown20;				//112B5
	UInt8				pciXState;				//112B6
	UInt8				connections;			//112B7
	
	UInt8				dispNum;				//112B8/11340
	ConnectionInfo		*connectionInfo;		//112BC, size 0x14
	UInt8				connectTaggedType;		//112BE
	UInt8				unknown21;				//112BF
	UInt8				connectTaggedData;		//112C0
	UInt8				unknown22;				//112C1
	UInt32				edidP1;					//112C4
	UInt32				edidP2;					//112C8
	UInt8				unknown23;				//112CC
	UInt8				edidP_lowByte;			//112D0
	CrtcValues			*scaledMode;			//112D4
												//112D8
	ushort				scaledFlags;			//112DC 12:dual-link, 
	ushort				currentClock;			//112DE
	ushort				unknown24;				//112E0
												//112E4
	ushort				unknown25;				//112E8
												//112EC
	void				*refCon;				//112F0
	IOFBInterruptProc	handlerFunction;		//112F4
	IOFBInterruptProc	enableFunction;			//112F8
	IOFBInterruptProc	disablerFunction;		//112FC
	InterruptSetID		setID1;					//11300
	InterruptMemberNumber member;				//11304
	InterruptSetID		setID2;					//11308
	//UInt8				activeFlags/timingFlags;//1130C, 4:CRT>5:CRT>6:LVDS>2:DFP1>3:DFP2>7:COMPTV>1:TV>0:NONE
	UInt8				connectedFlags;			//1130D
	UInt8				controlFlags;			//1130E
	UInt8				connectTypeFlags;		//1130F
	UInt8				edidFlags;				//11310
	DetailTimingState	DTState[4];				//11314, each size 0x24
	CrtcValues			detailModes[4];			//113A4, each size 1C
	ATIInfo				*atiInfo;				//11414, size 0x4C
	UInt8				edid[128];				//114A8/1153C
	UInt32				mirrorFlags;			//11528
	ScaledParameters	HWscaledPara;			//1152C,1152E
	ushort				nativeWidth;			//11530
	ushort				nativeHeight;			//11532
	ScaledParameters	SWscaledPara;			//11534,11536
	ATIFEDSInfo			*swScalerInfo;			//11538
	UInt32				rotate;					//1153C, 1:90 2:180 3:270
	UInt32				displayLineBuffer;		//11540
	UInt32				unknown26;				//11548
	UInt32				unknown27;				//1154C
	UInt8				htidFeature;			//11550
	UInt8				unknown28;				//11551
	UInt32				unknown29;				//11558
	UInt32				unknown30;				//1155C
	UInt32				unknown31;				//11560
	UInt32				unknown32;				//11564
	UInt32				unknown33;				//11568
	UInt32				unknown34;				//1156C
	UInt32				unknown35;				//11570
												//11594
												//1159C
												//115A4
	UInt8				unknown36;				//11621
	UInt8				unknown37;				//11622
	UInt8				unknown38;				//11623
	ATYSG				*atisg;					//11660 size 0x654
	IOService			*aProvider;				//11668
	IOService			*aDriver;				//1166C
	UInt32				videoClock;				//115BC
	UInt32				unknown39;				//115C0
	UInt32				unknown40;				//115C4
	UInt32				unknown41;				//115C8
												//11648
	UInt8				unknown42[13];			//1164C
	UInt32				intFlags[2];			//116CC
	UInt8				unknown43;				//11A06 wommy
} DriverGlobal;

#endif